@@ -16,6 +16,7 @@
 # limitations under the License.
 
 import struct
+import itertools
 
 from ryu import exception
 from ryu.lib import mac
@@ -39,9 +40,14 @@ FWW_NW_DSCP = 1 << 1
 FWW_NW_ECN = 1 << 2
 FWW_ARP_SHA = 1 << 3
 FWW_ARP_THA = 1 << 6
+FWW_IPV6_LABEL = 1 << 7
 FWW_NW_TTL = 1 << 8
 FWW_ALL = (1 << 13) - 1
 
+FLOW_NW_FRAG_ANY = 1 << 0
+FLOW_NW_FRAG_LATER = 1 << 1
+FLOW_NW_FRAG_MASK = FLOW_NW_FRAG_ANY | FLOW_NW_FRAG_LATER
+
 # Ethernet types, for set_dl_type()
 ETH_TYPE_IP = 0x0800
 ETH_TYPE_ARP = 0x0806
@@ -50,6 +56,7 @@ ETH_TYPE_IPV6 = 0x86dd
 ETH_TYPE_LACP = 0x8809
 
 IPPROTO_ICMP = 1
+IPPROTO_ICMPV6 = 58
 
 IP_ECN_MASK = 0x03
 IP_DSCP_MASK = 0xfc
@@ -62,6 +69,8 @@ MF_PACK_STRING_MAC = '!6s'
 
 _MF_FIELDS = {}
 
+FLOW_N_REGS = 8  # ovs 1.5
+
 
 class Flow(object):
     def __init__(self):
@@ -80,6 +89,14 @@ class Flow(object):
         self.nw_src = 0
         self.nw_dst = 0
         self.tun_id = 0
+        self.arp_spa = 0
+        self.arp_tpa = 0
+        self.ipv6_src = []
+        self.ipv6_dst = []
+        self.nd_target = []
+        self.nw_frag = 0
+        self.regs = [0] * FLOW_N_REGS
+        self.ipv6_label = 0
 
 
 class FlowWildcards(object):
@@ -91,7 +108,15 @@ class FlowWildcards(object):
         self.nw_src_mask = 0
         self.nw_dst_mask = 0
         self.tun_id_mask = 0
+        self.arp_spa_mask = 0
+        self.arp_tpa_mask = 0
         self.vlan_tci_mask = 0
+        self.ipv6_src_mask = []
+        self.ipv6_dst_mask = []
+        self.nd_target_mask = []
+        self.nw_frag_mask = 0
+        self.regs_bits = 0
+        self.regs_mask = [0] * FLOW_N_REGS
         self.wildcards = FWW_ALL
 
 
@@ -189,6 +214,28 @@ class ClsRule(object):
         self.wc.wildcards &= ~FWW_NW_TTL
         self.flow.nw_ttl = nw_ttl
 
+    def set_nw_frag(self, nw_frag):
+        self.wc.nw_frag_mask |= FLOW_NW_FRAG_MASK
+        self.flow.nw_frag = nw_frag
+
+    def set_nw_frag_masked(self, nw_frag, mask):
+        self.wc.nw_frag_mask = mask
+        self.flow.nw_frag = nw_frag & mask
+
+    def set_arp_spa(self, spa):
+        self.set_arp_spa_masked(spa, UINT32_MAX)
+
+    def set_arp_spa_masked(self, spa, mask):
+        self.flow.arp_spa = spa
+        self.wc.arp_spa_mask = mask
+
+    def set_arp_tpa(self, tpa):
+        self.set_arp_tpa_masked(tpa, UINT32_MAX)
+
+    def set_arp_tpa_masked(self, tpa, mask):
+        self.flow.arp_tpa = tpa
+        self.wc.arp_tpa_mask = mask
+
     def set_arp_sha(self, sha):
         self.wc.wildcards &= ~FWW_ARP_SHA
         self.flow.arp_sha = sha
@@ -197,6 +244,50 @@ class ClsRule(object):
         self.wc.wildcards &= ~FWW_ARP_THA
         self.flow.arp_tha = tha
 
+    def set_icmpv6_type(self, icmp_type):
+        self.set_tp_src(icmp_type)
+
+    def set_icmpv6_code(self, icmp_code):
+        self.set_tp_dst(icmp_code)
+
+    def set_ipv6_label(self, label):
+        self.wc.wildcards &= ~FWW_IPV6_LABEL
+        self.flow.ipv6_label = label
+
+    def set_ipv6_label(self, label):
+        self.wc.wildcards &= ~FWW_IPV6_LABEL
+        self.flow.ipv6_label = label
+
+    def set_ipv6_src_masked(self, src, mask):
+        self.wc.ipv6_src_mask = mask
+        self.flow.ipv6_src = [x & y for (x, y) in itertools.izip(src, mask)]
+
+    def set_ipv6_src(self, src):
+        self.flow.ipv6_src = src
+
+    def set_ipv6_dst_masked(self, dst, mask):
+        self.wc.ipv6_dst_mask = mask
+        self.flow.ipv6_dst = [x & y for (x, y) in itertools.izip(dst, mask)]
+
+    def set_ipv6_dst(self, dst):
+        self.flow.ipv6_dst = dst
+
+    def set_nd_target_masked(self, target, mask):
+        self.wc.nd_target_mask = mask
+        self.flow.nd_target = [x & y for (x, y) in
+                               itertools.izip(target, mask)]
+
+    def set_nd_target(self, target):
+        self.flow.nd_target = target
+
+    def set_reg(self, reg_idx, value):
+        self.set_reg_masked(reg_idx, value, 0)
+
+    def set_reg_masked(self, reg_idx, value, mask):
+        self.wc.regs_mask[reg_idx] = mask
+        self.flow.regs[reg_idx] = value
+        self.wc.regs_bits |= (1 << reg_idx)
+
     def flow_format(self):
         # Tunnel ID is only supported by NXM
         if self.wc.tun_id_mask != 0:
@@ -258,11 +349,12 @@ def mf_from_nxm_header(nxm_header):
         return None
     make = _MF_FIELDS.get(nxm_header)
     assert make is not None
-    return make()
+    return make(nxm_header)
 
 
 class MFField(object):
-    def __init__(self, pack_str):
+    def __init__(self, nxm_header, pack_str):
+        self.nxm_header = nxm_header
         self.pack_str = pack_str
         self.n_bytes = struct.calcsize(pack_str)
         self.n_bits = self.n_bytes * 8
@@ -273,7 +365,7 @@ class MFField(object):
 
     def putw(self, buf, offset, value, mask):
         len_ = self._put(buf, offset, value)
-        return len + self._put(buf, offset + len_, mask)
+        return len_ + self._put(buf, offset + len_, mask)
 
     def _is_all_ones(self, value):
         return value == (1 << self.n_bits) - 1
@@ -286,13 +378,24 @@ class MFField(object):
         else:
             return self.putw(buf, offset, value, mask)
 
+    def _putv6(self, buf, offset, value):
+        ofproto_parser.msg_pack_into(self.pack_str, buf, offset,
+                                     *value)
+        return self.n_bytes
+
+    def putv6(self, buf, offset, value, mask):
+        len_ = self._putv6(buf, offset, value)
+        if len(mask):
+            return len_ + self._putv6(buf, offset + len_, mask)
+        return len_
+
 
 @_register_make
 @_set_nxm_headers([ofproto_v1_0.NXM_OF_IN_PORT])
 class MFInPort(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_BE16)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE16)
 
     def put(self, buf, offset, rule):
         return self._put(buf, offset, rule.flow.in_port)
@@ -302,8 +405,8 @@ class MFInPort(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_OF_ETH_DST, ofproto_v1_0.NXM_OF_ETH_DST_W])
 class MFEthDst(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_MAC)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_MAC)
 
     def put(self, buf, offset, rule):
         if rule.wc.dl_dst_mask:
@@ -317,8 +420,8 @@ class MFEthDst(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_OF_ETH_SRC, ofproto_v1_0.NXM_OF_ETH_SRC_W])
 class MFEthSrc(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_MAC)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_MAC)
 
     def put(self, buf, offset, rule):
         if rule.wc.dl_src_mask:
@@ -332,8 +435,8 @@ class MFEthSrc(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_OF_ETH_TYPE])
 class MFEthType(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_BE16)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE16)
 
     def put(self, buf, offset, rule):
         return self._put(buf, offset, rule.flow.dl_type)
@@ -344,8 +447,8 @@ class MFEthType(MFField):
                    ofproto_v1_0.NXM_OF_VLAN_TCI_W])
 class MFVlan(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_BE16)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE16)
 
     def put(self, buf, offset, rule):
         return self.putm(buf, offset, rule.flow.vlan_tci,
@@ -356,8 +459,8 @@ class MFVlan(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_OF_IP_TOS])
 class MFIPDSCP(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_8)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_8)
 
     def put(self, buf, offset, rule):
         return self._put(buf, offset,
@@ -368,8 +471,8 @@ class MFIPDSCP(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_NX_TUN_ID, ofproto_v1_0.NXM_NX_TUN_ID_W])
 class MFTunId(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_BE64)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE64)
 
     def put(self, buf, offset, rule):
         return self.putm(buf, offset, rule.flow.tun_id, rule.wc.tun_id_mask)
@@ -379,8 +482,8 @@ class MFTunId(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_OF_IP_SRC, ofproto_v1_0.NXM_OF_IP_SRC_W])
 class MFIPSrc(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_BE32)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE32)
 
     def put(self, buf, offset, rule):
         return self.putm(buf, offset, rule.flow.nw_src, rule.wc.nw_src_mask)
@@ -390,8 +493,8 @@ class MFIPSrc(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_OF_IP_DST, ofproto_v1_0.NXM_OF_IP_DST_W])
 class MFIPDst(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_BE32)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE32)
 
     def put(self, buf, offset, rule):
         return self.putm(buf, offset, rule.flow.nw_dst, rule.wc.nw_dst_mask)
@@ -401,8 +504,8 @@ class MFIPDst(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_NX_IP_ECN])
 class MFIPECN(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_8)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_8)
 
     def put(self, buf, offset, rule):
         return self._put(buf, offset,
@@ -413,8 +516,8 @@ class MFIPECN(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_NX_IP_TTL])
 class MFIPTTL(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_8)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_8)
 
     def put(self, buf, offset, rule):
         return self._put(buf, offset, rule.flow.nw_ttl)
@@ -424,8 +527,8 @@ class MFIPTTL(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_OF_IP_PROTO])
 class MFIPProto(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_8)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_8)
 
     def put(self, buf, offset, rule):
         return self._put(buf, offset, rule.flow.nw_proto)
@@ -436,30 +539,110 @@ class MFIPProto(MFField):
                    ofproto_v1_0.NXM_OF_UDP_SRC, ofproto_v1_0.NXM_OF_UDP_SRC_W])
 class MFTPSRC(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_BE16)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE16)
 
     def put(self, buf, offset, rule):
         return self.putm(buf, offset, rule.flow.tp_src, rule.wc.tp_src_mask)
 
 
+@_register_make
+@_set_nxm_headers([ofproto_v1_0.NXM_OF_ARP_SPA, ofproto_v1_0.NXM_OF_ARP_SPA_W])
+class MFArpSpa(MFField):
+    @classmethod
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE32)
+
+    def put(self, buf, offset, rule):
+        return self.putm(buf, offset, rule.flow.arp_spa, rule.wc.arp_spa_mask)
+
+
+@_register_make
+@_set_nxm_headers([ofproto_v1_0.NXM_OF_ARP_TPA, ofproto_v1_0.NXM_OF_ARP_TPA_W])
+class MFArpTpa(MFField):
+    @classmethod
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE32)
+
+    def put(self, buf, offset, rule):
+        return self.putm(buf, offset, rule.flow.arp_tpa, rule.wc.arp_tpa_mask)
+
+
 @_register_make
 @_set_nxm_headers([ofproto_v1_0.NXM_NX_ARP_SHA])
 class MFArpSha(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_MAC)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_MAC)
 
     def put(self, buf, offset, rule):
         return self._put(buf, offset, rule.flow.arp_sha)
 
 
+@_register_make
+@_set_nxm_headers([ofproto_v1_0.NXM_NX_IPV6_SRC,
+                   ofproto_v1_0.NXM_NX_IPV6_SRC_W])
+class MFIPV6Src(MFField):
+    @classmethod
+    def make(cls, header):
+        return cls(header, '!4I')
+
+    def put(self, buf, offset, rule):
+        return self.putv6(buf, offset,
+                          rule.flow.ipv6_src,
+                          rule.wc.ipv6_src_mask)
+
+
+@_register_make
+@_set_nxm_headers([ofproto_v1_0.NXM_NX_IPV6_DST,
+                   ofproto_v1_0.NXM_NX_IPV6_DST_W])
+class MFIPV6Dst(MFField):
+    @classmethod
+    def make(cls, header):
+        return cls(header, '!4I')
+
+    def put(self, buf, offset, rule):
+        return self.putv6(buf, offset,
+                          rule.flow.ipv6_dst,
+                          rule.wc.ipv6_dst_mask)
+
+
+@_register_make
+@_set_nxm_headers([ofproto_v1_0.NXM_NX_ND_TARGET,
+                   ofproto_v1_0.NXM_NX_ND_TARGET_W])
+class MFNdTarget(MFField):
+    @classmethod
+    def make(cls, header):
+        return cls(header, '!4I')
+
+    def put(self, buf, offset, rule):
+        return self.putv6(buf, offset,
+                          rule.flow.nd_target,
+                          rule.wc.nd_target_mask)
+
+
+@_register_make
+@_set_nxm_headers([ofproto_v1_0.NXM_NX_IP_FRAG,
+                   ofproto_v1_0.NXM_NX_IP_FRAG_W])
+class MFIpFrag(MFField):
+    @classmethod
+    def make(cls, header):
+        return cls(header, '!B')
+
+    def put(self, buf, offset, rule):
+        if rule.wc.nw_frag_mask == FLOW_NW_FRAG_MASK:
+            return self._put(buf, offset, rule.flow.nw_frag)
+        else:
+            return self.putw(buf, offset, rule.flow.nw_frag,
+                             rule.wc.nw_frag_mask & FLOW_NW_FRAG_MASK)
+
+
 @_register_make
 @_set_nxm_headers([ofproto_v1_0.NXM_NX_ARP_THA])
 class MFArpTha(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_MAC)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_MAC)
 
     def put(self, buf, offset, rule):
         return self._put(buf, offset, rule.flow.arp_tha)
@@ -469,22 +652,74 @@ class MFArpTha(MFField):
 @_set_nxm_headers([ofproto_v1_0.NXM_OF_ICMP_TYPE])
 class MFICMPType(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_8)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_8)
 
     def put(self, buf, offset, rule):
-        return self._put(buf, offset, rule.flow.nw_src)
+        return self._put(buf, offset, rule.flow.tp_src)
 
 
 @_register_make
 @_set_nxm_headers([ofproto_v1_0.NXM_OF_ICMP_CODE])
 class MFICMPCode(MFField):
     @classmethod
-    def make(cls):
-        return cls(MF_PACK_STRING_8)
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_8)
 
     def put(self, buf, offset, rule):
-        return self._put(buf, offset, rule.flow.nw_dst)
+        return self._put(buf, offset, rule.flow.tp_dst)
+
+
+@_register_make
+@_set_nxm_headers([ofproto_v1_0.NXM_NX_ICMPV6_TYPE])
+class MFICMPV6Type(MFField):
+    @classmethod
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_8)
+
+    def put(self, buf, offset, rule):
+        return self._put(buf, offset, rule.flow.tp_src)
+
+
+@_register_make
+@_set_nxm_headers([ofproto_v1_0.NXM_NX_ICMPV6_CODE])
+class MFICMPV6Code(MFField):
+    @classmethod
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_8)
+
+    def put(self, buf, offset, rule):
+        return self._put(buf, offset, rule.flow.tp_dst)
+
+
+@_register_make
+@_set_nxm_headers([ofproto_v1_0.NXM_NX_IPV6_LABEL])
+class MFICMPV6Label(MFField):
+    @classmethod
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE32)
+
+    def put(self, buf, offset, rule):
+        return self._put(buf, offset, rule.flow.ipv6_label)
+
+
+@_register_make
+@_set_nxm_headers([ofproto_v1_0.nxm_nx_reg(i) for i in range(FLOW_N_REGS)]
+                  + [ofproto_v1_0.nxm_nx_reg_w(i) for i in range(FLOW_N_REGS)])
+class MFRegister(MFField):
+    @classmethod
+    def make(cls, header):
+        return cls(header, MF_PACK_STRING_BE32)
+
+    def put(self, buf, offset, rule):
+        for i in range(FLOW_N_REGS):
+            if (ofproto_v1_0.nxm_nx_reg(i) == self.nxm_header or
+                ofproto_v1_0.nxm_nx_reg_w(i) == self.nxm_header):
+                if rule.wc.regs_mask[i]:
+                    return self.putm(buf, offset, rule.flow.regs[i],
+                                     rule.wc.regs_mask[i])
+                else:
+                    return self._put(buf, offset, rule.flow.regs[i])
 
 
 def serialize_nxm_match(rule, buf, offset):
@@ -533,10 +768,8 @@ def serialize_nxm_match(rule, buf, offset):
                                                  == IPPROTO_ICMP):
         if rule.wc.tp_src_mask != 0:
             offset += nxm_put(buf, offset, ofproto_v1_0.NXM_OF_ICMP_TYPE, rule)
-        elif rule.wc.tp_dst_mask != 0:
+        if rule.wc.tp_dst_mask != 0:
             offset += nxm_put(buf, offset, ofproto_v1_0.NXM_OF_ICMP_CODE, rule)
-        else:
-            pass
 
     if rule.flow.tp_src != 0:
         if rule.flow.nw_proto == 6:
@@ -585,13 +818,67 @@ def serialize_nxm_match(rule, buf, offset):
             header = ofproto_v1_0.NXM_OF_IP_DST_W
         offset += nxm_put(buf, offset, header, rule)
 
-    # XXX: IPv6
+    # IPv6
+    if not rule.wc.wildcards & FWW_NW_PROTO and (rule.flow.nw_proto
+                                                 == IPPROTO_ICMPV6):
+        if rule.wc.tp_src_mask != 0:
+            offset += nxm_put(buf, offset, ofproto_v1_0.NXM_NX_ICMPV6_TYPE,
+                              rule)
+        if rule.wc.tp_dst_mask != 0:
+            offset += nxm_put(buf, offset, ofproto_v1_0.NXM_NX_ICMPV6_CODE,
+                              rule)
+
+    if not rule.wc.wildcards & FWW_IPV6_LABEL:
+        offset += nxm_put(buf, offset, ofproto_v1_0.NXM_NX_IPV6_LABEL, rule)
+
+    if len(rule.flow.ipv6_src):
+        if len(rule.wc.ipv6_src_mask):
+            header = ofproto_v1_0.NXM_NX_IPV6_SRC_W
+        else:
+            header = ofproto_v1_0.NXM_NX_IPV6_SRC
+        offset += nxm_put(buf, offset, header, rule)
+
+    if len(rule.flow.ipv6_dst):
+        if len(rule.wc.ipv6_dst_mask):
+            header = ofproto_v1_0.NXM_NX_IPV6_DST_W
+        else:
+            header = ofproto_v1_0.NXM_NX_IPV6_DST
+        offset += nxm_put(buf, offset, header, rule)
+
+    if len(rule.flow.nd_target):
+        if len(rule.wc.nd_target_mask):
+            header = ofproto_v1_0.NXM_NX_ND_TARGET_W
+        else:
+            header = ofproto_v1_0.NXM_NX_ND_TARGET
+        offset += nxm_put(buf, offset, header, rule)
+
     # ARP
+    if rule.flow.arp_spa != 0:
+        if rule.wc.arp_spa_mask == UINT32_MAX:
+            header = ofproto_v1_0.NXM_OF_ARP_SPA
+        else:
+            header = ofproto_v1_0.NXM_OF_ARP_SPA_W
+        offset += nxm_put(buf, offset, header, rule)
+
+    if rule.flow.arp_tpa != 0:
+        if rule.wc.arp_tpa_mask == UINT32_MAX:
+            header = ofproto_v1_0.NXM_OF_ARP_TPA
+        else:
+            header = ofproto_v1_0.NXM_OF_ARP_TPA_W
+        offset += nxm_put(buf, offset, header, rule)
+
     if not rule.wc.wildcards & FWW_ARP_SHA:
         offset += nxm_put(buf, offset, ofproto_v1_0.NXM_NX_ARP_SHA, rule)
     if not rule.wc.wildcards & FWW_ARP_THA:
         offset += nxm_put(buf, offset, ofproto_v1_0.NXM_NX_ARP_THA, rule)
 
+    if rule.flow.nw_frag:
+        if rule.wc.nw_frag_mask == FLOW_NW_FRAG_MASK:
+            header = ofproto_v1_0.NXM_NX_IP_FRAG
+        else:
+            header = ofproto_v1_0.NXM_NX_IP_FRAG_W
+        offset += nxm_put(buf, offset, header, rule)
+
     # Tunnel Id
     if rule.wc.tun_id_mask != 0:
         if rule.wc.tun_id_mask == UINT64_MAX:
@@ -602,6 +889,14 @@ def serialize_nxm_match(rule, buf, offset):
 
     # XXX: Cookie
 
+    for i in range(FLOW_N_REGS):
+        if rule.wc.regs_bits & (1 << i):
+            if rule.wc.regs_mask[i]:
+                header = ofproto_v1_0.nxm_nx_reg_w(i)
+            else:
+                header = ofproto_v1_0.nxm_nx_reg(i)
+            offset += nxm_put(buf, offset, header, rule)
+
     # Pad
     pad_len = round_up(offset) - offset
     ofproto_parser.msg_pack_into("%dx" % pad_len, buf, offset)
@@ -614,7 +909,7 @@ def nxm_put(buf, offset, header, rule):
     nxm = NXMatch(header)
     len_ = nxm.put_header(buf, offset)
     mf = mf_from_nxm_header(nxm.header)
-    return len_ + mf.put(buf, offset + len, rule)
+    return len_ + mf.put(buf, offset + len_, rule)
 
 
 def round_up(length):
